﻿/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2020 NiuTrans Research. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com) 2020-08-09
 * $Updated by: CAO Hang and Wu Siming 2020-12-13
 */

#include <cstdlib>
#include <algorithm>

#include "TrainDataSet.h"
#include "../Utility.h"
#include "../translate/Vocab.h"

using namespace nmt;

namespace nts {

/* get the maximum source sentence length in a range */
int TrainDataSet::MaxSrcLen(XList* buf, int begin, int end) {
    CheckNTErrors((end > begin) && (begin >= 0) && (end <= buf->count), "Invalid range");
    int maxLen = 0;
    for (int i = begin; i < end; i++) {
        IntList* srcSent = ((TrainExample*)buf->Get(i))->srcSent;
        maxLen = MAX(int(srcSent->Size()), maxLen);
    }
    return maxLen;
}

/* get the maximum target sentence length in a range */
int TrainDataSet::MaxTgtLen(XList* buf, int begin, int end) {
    CheckNTErrors((end > begin) && (begin >= 0) && (end <= buf->count), "Invalid range");
    int maxLen = 0;
    for (int i = begin; i < end; i++) {
        IntList* tgtSent = ((TrainExample*)buf->Get(i))->tgtSent;
        maxLen = MAX(int(tgtSent->Size()), maxLen);
    }
    return maxLen;
}

/* sort the buffer by source sentence length (in descending order) */
void TrainDataSet::SortBySrcLength(XList* buf) {
    stable_sort(buf->items, buf->items + buf->count,
                [](void* a, void* b) {
                    return ((TrainExample*)(a))->srcSent->Size() <
                           ((TrainExample*)(b))->srcSent->Size();
                });
}
/* sort the buffer by target sentence length (in descending order) */
void TrainDataSet::SortByTgtLength(XList* buf) {
    stable_sort(buf->items, buf->items + buf->count, 
                [](void* a, void* b) {
                    return ((TrainExample*)(a))->tgtSent->Size() < 
                           ((TrainExample*)(b))->tgtSent->Size();
                });
}

/* sort buckets by key (in descending order) */
void TrainDataSet::SortBuckets(XList* buf) {
    sort(buf->items, buf->items + buf->count,
         [](void* a, void* b) {
             return ((TrainExample*)(a))->bucketKey <
                    ((TrainExample*)(b))->bucketKey;
         });
}

/*
load samples from a file into the buffer
training data format (binary):
first 8 bit: number of sentence pairs
subsequent segements:
source sentence length (4 bit)
target sentence length (4 bit)
source tokens (4 bit per token)
target tokens (4 bit per token)
>> buf - the buffer (list) of samples 
*/
bool TrainDataSet::LoadBatchToBuf(XList* buf)
{
    ClearSamples(buf);

    int sampleNum = 0;

    while ((sampleNum < bufferSize)) {

        int srcLen = 0;
        int tgtLen = 0;
        
        size_t n = fread(&srcLen, sizeof(int), 1, fp);
        if (n == 0)
            break;

        fread(&tgtLen, sizeof(int), 1, fp);
        CheckNTErrors(srcLen > 0, "Invalid source sentence length");
        CheckNTErrors(tgtLen > 0, "Invalid target sentence length");

        IntList *srcSent = new IntList(srcLen);
        IntList *tgtSent = new IntList(tgtLen);
        srcSent->ReadFromFile(fp, srcLen);
        tgtSent->ReadFromFile(fp, tgtLen);

        TrainExample* example = new TrainExample(sampleNum++, 0, srcSent, tgtSent);
        buf->Add(example);
    }

    /* reset the file pointer to the begin */
    if (feof(fp) && isTraining) {

        rewind(fp);

        int srcVocabSize = 0;
        int tgtVocabSize = 0;
        fread(&srcVocabSize, sizeof(int), 1, fp);
        fread(&tgtVocabSize, sizeof(int), 1, fp);
        fread(&totalSampleNum, sizeof(totalSampleNum), 1, fp);
    }

    /* group samples in the buffer into buckets */
    SortByTgtLength(buf);

    SortBySrcLength(buf);

    if (isTraining)
        BuildBucket(buf);

    return true;
}

/*
load a mini-batch to a device
>> buf - the buffer (list) of samples
>> curIdx - the index of the buffer
>> batchEnc - a tensor to store the batch of encoder input
>> paddingEnc - a tensor to store the batch of encoder paddings
>> batchDec - a tensor to store the batch of decoder input
>> paddingDec - a tensor to store the batch of decoder paddings
>> label - a tensor to store the label of input
>> minSentBatch - the minimum number of sentence batch
>> batchSize - the maxium number of words in a batch
>> devID - the device id, -1 for the CPU
>> wc - number of target words in a batch
>> sc - number of samples in a batch
*/
bool TrainDataSet::LoadBatch(XList* buf, int & curIdx,
                             XTensor* batchEnc, XTensor* paddingEnc,
                             XTensor* batchDec, XTensor* paddingDec, XTensor* label, 
                             int minSentBatch, int batchSize, int devID, int& wc, int& sc)
{
    int srcTokenNum = 0;
    int tgtTokenNum = 0;
    int realBatchSize = 0;

    /* dynamic batching for sentences */
    int bucketKey = ((TrainExample*)(buf->Get(curIdx)))->bucketKey;
    while ((realBatchSize < (int(buf->Size()) - curIdx)) &&
        (((TrainExample*)(buf->Get(curIdx + realBatchSize)))->bucketKey == bucketKey)) {
        realBatchSize++;
    }

    realBatchSize = MIN(realBatchSize, (int(buf->Size()) - curIdx));
    CheckNTErrors(realBatchSize > 0, "Invalid batch size");

    /* get the maximum target sentence length in a mini-batch */
    int maxSrcLen = MaxSrcLen(buf, curIdx, curIdx + realBatchSize);
    int maxTgtLen = MaxTgtLen(buf, curIdx, curIdx + realBatchSize);

    CheckNTErrors(maxSrcLen > 0, "Invalid source length for batching");
    CheckNTErrors(maxTgtLen > 0, "Invalid target length for batching");

    int* batchEncValues = new int[realBatchSize * maxSrcLen];
    float* paddingEncValues = new float[realBatchSize * maxSrcLen];

    int* labelVaues = new int[realBatchSize * maxTgtLen];
    int* batchDecValues = new int[realBatchSize * maxTgtLen];
    float* paddingDecValues = new float[realBatchSize * maxTgtLen];

    for (int i = 0; i < realBatchSize * maxSrcLen; i++) {
        batchEncValues[i] = 1;
        paddingEncValues[i] = 1.0F;
    }
    for (int i = 0; i < realBatchSize * maxTgtLen; i++) {
        batchDecValues[i] = 1;
        labelVaues[i] = 1;
        paddingDecValues[i] = 1.0F;
    }

    int curSrc = 0;
    int curTgt = 0;

    /*
    batchEnc: end with EOS (left padding)
    batchDec: begin with SOS (right padding)
    label:    end with EOS (right padding)
    */
    for (int i = 0; i < realBatchSize; ++i) {

        TrainExample* sample = (TrainExample*)(buf->Get(curIdx + i));
        srcTokenNum += int(sample->srcSent->Size());
        tgtTokenNum += int(sample->tgtSent->Size());

        curSrc = maxSrcLen * i;
        for (int j = 0; j < sample->srcSent->Size(); j++) {
            batchEncValues[curSrc++] = sample->srcSent->Get(j);
        }

        curTgt = maxTgtLen * i;
        for (int j = 0; j < sample->tgtSent->Size(); j++) {
            if (j > 0)
                labelVaues[curTgt - 1] = sample->tgtSent->Get(j);
            batchDecValues[curTgt++] = sample->tgtSent->Get(j);
        }
        labelVaues[curTgt - 1] = 2;
        while (curSrc < maxSrcLen * (i + 1))
            paddingEncValues[curSrc++] = 0;
        while (curTgt < maxTgtLen * (i + 1))
            paddingDecValues[curTgt++] = 0;
    }

    InitTensor2D(batchEnc, realBatchSize, maxSrcLen, X_INT, devID);
    InitTensor2D(paddingEnc, realBatchSize, maxSrcLen, X_FLOAT, devID);
    InitTensor2D(batchDec, realBatchSize, maxTgtLen, X_INT, devID);
    InitTensor2D(paddingDec, realBatchSize, maxTgtLen, X_FLOAT, devID);
    InitTensor2D(label, realBatchSize, maxTgtLen, X_INT, devID);

    curIdx += realBatchSize;

    batchEnc->SetData(batchEncValues, batchEnc->unitNum);
    paddingEnc->SetData(paddingEncValues, paddingEnc->unitNum);
    batchDec->SetData(batchDecValues, batchDec->unitNum);
    paddingDec->SetData(paddingDecValues, paddingDec->unitNum);
    label->SetData(labelVaues, label->unitNum);

    delete[] batchEncValues;
    delete[] paddingEncValues;
    delete[] batchDecValues;
    delete[] paddingDecValues;
    delete[] labelVaues;

    wc = tgtTokenNum;
    sc = realBatchSize;
    return true;
}

/* 
clear the buffer 
>> buf - the buffer (list) of samples
*/
void TrainDataSet::ClearSamples(XList* buf)
{
    for (int i = 0; i < buf->count; i++) {
        TrainExample* sample = (TrainExample*)buf->Get(i);
        delete sample;
    }
    buf->Clear();
}

/*
the constructor of DataSet
>> dataFile - path of the data file
>> bucketSize - size of the bucket to keep similar length sentence pairs
>> training - indicates whether it is used for training
*/
void TrainDataSet::Init(const char* dataFile, int myBucketSize, bool training)
{
    fp = fopen(dataFile, "rb");
    CheckNTErrors(fp, "can not open the training file");

    int srcVocabSize = 0;
    int tgtVocabSize = 0;
    fread(&srcVocabSize, sizeof(int), 1, fp);
    fread(&tgtVocabSize, sizeof(int), 1, fp);
    CheckNTErrors(srcVocabSize > 0, "Invalid source vocabulary size");
    CheckNTErrors(tgtVocabSize > 0, "Invalid target vocabulary size");

    fread(&totalSampleNum, sizeof(totalSampleNum), 1, fp);
    CheckNTErrors(totalSampleNum > 0, "Invalid sentence pairs number");

    bucketSize = myBucketSize;
    isTraining = training;
}

/* group data with similar length into buckets */
void TrainDataSet::BuildBucket(XList * buf)
{
    int idx = 0;

    /* build buckets by the length of source and target sentences */
    while (idx < int(buf->Size())) {

        /* sentence number in a bucket */
        int sentNum = 1;

        /* get the maximum source sentence length in a bucket */
        int maxSrcLen = MaxSrcLen(buf, idx, idx + sentNum);
        int maxTgtLen = MaxTgtLen(buf, idx, idx + sentNum);
        int maxLen = MAX(maxSrcLen, maxTgtLen);

        /* the maximum sentence number in a bucket */
        const int MAX_SENT_NUM = 5120;

        while ((sentNum < (buf->count - idx))
            && (sentNum < MAX_SENT_NUM)
            && (sentNum * maxLen <= bucketSize)) {
            sentNum++;
            maxSrcLen = MaxSrcLen(buf, idx, idx + sentNum);
            maxTgtLen = MaxTgtLen(buf, idx, idx + sentNum);
            maxLen = MAX(maxSrcLen, maxTgtLen);
        }

        /* make sure the number is valid */
        if ((sentNum) * maxLen > bucketSize || sentNum >= MAX_SENT_NUM) {
            sentNum--;
            sentNum = max(8 * (sentNum / 8), sentNum % 8);
        }
        if ((int(buf->Size()) - idx) < sentNum)
            sentNum = int(buf->Size()) - idx;

        /* assign the same key for items in a bucket */
        int randomKey = rand();
        for (int i = 0; i < sentNum; i++) {
            ((TrainExample*)(buf->Get(idx + i)))->bucketKey = randomKey;
        }

        idx += sentNum;
    }

    /* sort buckets by their keys */
    SortBuckets(buf);
}

/* de-constructor */
TrainDataSet::~TrainDataSet()
{
    fclose(fp);
}

/* constructor */
TrainExample::TrainExample(int myID, int myKey, IntList* s, IntList* t)
{
    id = myID;
    bucketKey = myKey;
    srcSent = s;
    tgtSent = t;
}

/* de-constructor */
TrainExample::~TrainExample()
{
    delete srcSent;
    delete tgtSent;
}

}
