/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2020 NiuTrans Research. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: HU Chi (huchinlp@foxmail.com) 2019-04-03
 * $Modified by: HU Chi (huchinlp@gmail.com) 2020-06
 */

#ifndef __TRAIN_DATASET_H__
#define __TRAIN_DATASET_H__

#include <cstdio>
#include <vector>
#include <fstream>

#include "../../../tensor/XList.h"
#include "../../../tensor/XTensor.h"
#include "../../../tensor/XGlobal.h"


using namespace std;

namespace nts {

/* a class of sentence pairs for training */
struct TrainExample {

public:
    /* id of the sentence pair */
    int id;

    /* source language setence (tokenized) */
    IntList * srcSent;

    /* target language setence (tokenized) */
    IntList * tgtSent;

    /* the key used to shuffle buckets */
    int bucketKey;

public:
    /* constructor */
    TrainExample(int myID, int myKey, IntList* s, IntList* t);

    /* de-constructor */
    ~TrainExample();
};

struct ReservedIDs {
    /* the padding id */
    int padID;

    /* the unk id */
    int unkID;

    /* start symbol */
    int startID;

    /* end symbol */
    int endID;
};

/* A `TrainDataSet` is associated with a file which contains training data. */
struct TrainDataSet {

public:

    /* the pointer to file stream */
    FILE* fp;

    /* number of training samples */
    size_t totalSampleNum;

    /* buffer size */
    size_t bufferSize;

    /* size of the bucket used for grouping sentences */
    size_t bucketSize;

    /* indicates whether it is used for training */
    bool isTraining;

    /* the padding id */
    int padID;

    /* the unk id */
    int unkID;

    /* start symbol */
    int startID;

    /* end symbol */
    int endID;

    /* the maximum length for a source sentence */
    int maxSrcLen;

    /* the maximum length for a target sentence */
    int maxTgtLen;

public:

    /* get the maximum source sentence length in a range */
    static
    int MaxSrcLen(XList* buf, int begin, int end);

    /* get the maximum target sentence length in a range */
    static
    int MaxTgtLen(XList* buf, int begin, int end);

    /* sort the input by source sentence length (in descending order) */
    void SortBySrcLength(XList* buf);

    /* sort the input by target sentence length (in descending order) */
    void SortByTgtLength(XList* buf);

    /* sort buckets by key (in descending order) */
    void SortBuckets(XList* buf);

    /* load the samples into the buffer (a list) */
    bool LoadBatchToBuf(XList * buf);

    /* load the samples into tensors from the buffer */
    static
    bool LoadBatch(XList * buf, int & curIdx,
                   XTensor* batchEnc, XTensor* paddingEnc,
                   XTensor* batchDec, XTensor* paddingDec, XTensor* label, int minSentBatch, int batchSize, int devID,
                   int &wc, int &sc);

    /* release the samples in a buffer */
    static
    void ClearSamples(XList * buf);

    /* initialization function */
    void Init(const char* dataFile, int bucketSize, bool training);

    /* group data into buckets with similar length */
    void BuildBucket(XList * buf);

    /* de-constructor */
    ~TrainDataSet();
};
}

#endif // __TRAIN_DATASET_H__